@@ -1,45 +1,45 @@
 from __future__ import annotations
 
 import sys
-
+from contextlib import AbstractContextManager
 from copy import deepcopy
-from typing import TYPE_CHECKING, Any, TypeVar
+from types import TracebackType
+from typing import TYPE_CHECKING, Any, Generic, TypeVar
 
 from typing_extensions import Self
 
-from ..types import MISSING, FuncExceptT, SupportsString, MissingT
+from ..types import MISSING, FuncExceptT, MissingT, SupportsString
 
 __all__ = [
-    'CustomError',
-
-    'CustomValueError',
-    'CustomIndexError',
-    'CustomOverflowError',
-    'CustomKeyError',
-    'CustomTypeError',
-    'CustomRuntimeError',
-    'CustomNotImplementedError',
-    'CustomPermissionError'
+    "CustomError",
+    "CustomIndexError",
+    "CustomKeyError",
+    "CustomNotImplementedError",
+    "CustomOverflowError",
+    "CustomPermissionError",
+    "CustomRuntimeError",
+    "CustomTypeError",
+    "CustomValueError",
 ]
 
 
 if TYPE_CHECKING:
-    class ExceptionT(Exception):
+
+    class ExceptionError(Exception):
         __name__: str
         __qualname__: str
-        ...
 else:
-    ExceptionT = Exception
+    ExceptionError = Exception
 
 
 class CustomErrorMeta(type):
     """Custom base exception meta class."""
 
     def __new__(cls: type[SelfCErrorMeta], *args: Any) -> SelfCErrorMeta:
-        return CustomErrorMeta.setup_exception(type.__new__(cls, *args))  # type: ignore
+        return CustomErrorMeta.setup_exception(super().__new__(cls, *args))
 
     @staticmethod
-    def setup_exception(exception: SelfCErrorMeta, override: str | ExceptionT | None = None) -> SelfCErrorMeta:
+    def setup_exception(exception: SelfCErrorMeta, override: str | ExceptionError | None = None) -> SelfCErrorMeta:
         """
         Setup an exception for later use in CustomError.
 
@@ -55,28 +55,28 @@ class CustomErrorMeta(type):
             else:
                 over_name, over_qual = override.__name__, override.__qualname__
 
-            if over_name.startswith('Custom'):
+            if over_name.startswith("Custom"):
                 exception.__name__ = over_name
             else:
-                exception.__name__ = f'Custom{over_name}'
+                exception.__name__ = f"Custom{over_name}"
 
             exception.__qualname__ = over_qual
 
-        if exception.__qualname__.startswith('Custom'):
+        if exception.__qualname__.startswith("Custom"):
             exception.__qualname__ = exception.__qualname__[6:]
 
         if sys.stdout and sys.stdout.isatty():
-            exception.__qualname__ = f'\033[0;31;1m{exception.__qualname__}\033[0m'
+            exception.__qualname__ = f"\033[0;31;1m{exception.__qualname__}\033[0m"
 
         exception.__module__ = Exception.__module__
 
         return exception
 
 
-SelfCErrorMeta = TypeVar('SelfCErrorMeta', bound=CustomErrorMeta)
+SelfCErrorMeta = TypeVar("SelfCErrorMeta", bound=CustomErrorMeta)
 
 
-class CustomError(ExceptionT, metaclass=CustomErrorMeta):
+class CustomError(ExceptionError, metaclass=CustomErrorMeta):
     """Custom base exception class."""
 
     def __init__(
@@ -97,7 +97,7 @@ class CustomError(ExceptionT, metaclass=CustomErrorMeta):
 
         super().__init__(message)
 
-    def __class_getitem__(cls, exception: str | type[ExceptionT] | ExceptionT) -> CustomError:
+    def __class_getitem__(cls, exception: str | type[ExceptionError] | ExceptionError) -> CustomError:
         from warnings import warn
 
         warn("Custom error is not subscriptable anymore. Don't use it", DeprecationWarning)
@@ -109,7 +109,7 @@ class CustomError(ExceptionT, metaclass=CustomErrorMeta):
         message: SupportsString | None | MissingT = MISSING,
         func: FuncExceptT | None | MissingT = MISSING,
         reason: SupportsString | FuncExceptT | None | MissingT = MISSING,
-        **kwargs: Any
+        **kwargs: Any,
     ) -> Self:
         """
         Copy an existing exception with defaults and instantiate a new one.
@@ -140,40 +140,95 @@ class CustomError(ExceptionT, metaclass=CustomErrorMeta):
         message = self.message
 
         if not message:
-            message = 'An error occurred!'
+            message = "An error occurred!"
 
         if self.func:
             func_header = norm_func_name(self.func).strip()
 
             if sys.stdout and sys.stdout.isatty():
-                func_header = f'\033[0;36m{func_header}\033[0m'
+                func_header = f"\033[0;36m{func_header}\033[0m"
 
-            func_header = f'({func_header}) '
+            func_header = f"({func_header}) "
         else:
-            func_header = ''
+            func_header = ""
 
         if self.kwargs:
-            self.kwargs = {
-                key: norm_display_name(value) for key, value in self.kwargs.items()
-            }
+            self.kwargs = {key: norm_display_name(value) for key, value in self.kwargs.items()}
 
         if self.reason:
             reason = self.reason = norm_display_name(self.reason)
 
             if reason:
                 if not isinstance(self.reason, dict):
-                    reason = f'({reason})'
+                    reason = f"({reason})"
 
                 if sys.stdout and sys.stdout.isatty():
-                    reason = f'\033[0;33m{reason}\033[0m'
-                reason = f' {reason}'
+                    reason = f"\033[0;33m{reason}\033[0m"
+                reason = f" {reason}"
         else:
-            reason = ''
+            reason = ""
 
-        return f'{func_header}{self.message!s}{reason}'.format(**self.kwargs).strip()
+        out = f"{func_header}{self.message!s}{reason}".format(**self.kwargs).strip()
+
+        if sys.version_info < (3, 13) and hasattr(self, "__notes__"):
+            out += "\n" + "\n".join(self.__notes__)
+
+        return out
+
+    if sys.version_info < (3, 11):
+        __notes__: list[str]
+
+        def add_note(self, note: str, /) -> None:
+            if not hasattr(self, "__notes__"):
+                self.__notes__ = []
+
+            self.__notes__.append(note)
+
+    @classmethod
+    def catch(cls) -> CatchError[Self]:
+        """
+        Create a context manager that catches exceptions of this class type.
+
+        Returns:
+            CatchError[Self]: A context manager that will catch and store exceptions of type `cls`
+                when used in a `with` block.
+        """
+        return CatchError(cls)
 
 
-SelfError = TypeVar('SelfError', bound=CustomError)
+SelfError = TypeVar("SelfError", bound=CustomError)
+
+
+class CatchError(AbstractContextManager["CatchError[SelfError]"], Generic[SelfError]):
+    """
+    Context manager for catching a specific exception type.
+    """
+
+    error: SelfError | None
+    """The caught exception instance, if any."""
+    tb: TracebackType | None
+    """The traceback object associated with the caught exception."""
+
+    def __init__(self, error: type[SelfError]) -> None:
+        self.error = None
+        self.tb = None
+        self._to_catch_error = error
+
+    def __enter__(self) -> Self:
+        return self
+
+    def __exit__(
+        self,
+        exc_type: type[BaseException] | None,
+        exc_value: BaseException | None,
+        traceback: TracebackType | None,
+    ) -> bool | None:
+        if isinstance(exc_value, self._to_catch_error):
+            self.error = exc_value
+            self.tb = traceback
+            return True
+
+        return None
 
 
 class CustomValueError(CustomError, ValueError):
