@@ -1,41 +1,51 @@
 from __future__ import annotations
 
-from functools import partial, wraps
+from contextlib import suppress
+from functools import wraps
 from inspect import Signature
 from inspect import _empty as empty_param
 from typing import (
-    TYPE_CHECKING, Any, Callable, Concatenate, Generator, Generic, Iterable, Iterator, Mapping, NoReturn, Protocol,
-    Sequence, TypeVar, cast, overload
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    ClassVar,
+    Concatenate,
+    Generator,
+    Generic,
+    Iterable,
+    Iterator,
+    Mapping,
+    NoReturn,
+    Protocol,
+    Sequence,
+    cast,
+    overload,
 )
 
-from typing_extensions import Self
+from typing_extensions import Self, deprecated
 
-from .builtins import F0, F1, P0, P1, R0, R1, T0, T1, T2, KwargsT, P, R, R0_co, R1_co, R_co, T, T0_co, T1_co, T_co
+from .builtins import F0, F1, P0, P1, R0, R1, T0, KwargsT, P, R, R0_co, R1_co, R_co, T, T0_co, T1_co, T_co
 
 __all__ = [
-    'copy_signature',
-
-    'inject_self',
-
-    'inject_kwargs_params',
-
-    'complex_hash',
-
-    'get_subclasses',
-
-    'classproperty', 'cachedproperty',
-
-    'KwargsNotNone',
-
-    'Singleton', 'to_singleton',
-
-    'LinearRangeLut'
+    "KwargsNotNone",
+    "LinearRangeLut",
+    "Singleton",
+    "cachedproperty",
+    "classproperty",
+    "complex_hash",
+    "copy_signature",
+    "get_subclasses",
+    "inject_kwargs_params",
+    "inject_self",
+    "to_singleton",
 ]
+# ruff: noqa: N801
 
 
 class copy_signature(Generic[F0]):
     """
-    Type util to copy the signature of one function to another function.\n
+    Type util to copy the signature of one function to another function.
+
     Especially useful for passthrough functions.
 
     .. code-block::
@@ -70,8 +80,7 @@ class copy_signature(Generic[F0]):
 class injected_self_func(Protocol[T_co, P, R_co]):
     @overload
     @staticmethod
-    def __call__(*args: P.args, **kwargs: P.kwargs) -> R_co:
-        ...
+    def __call__(*args: P.args, **kwargs: P.kwargs) -> R_co: ...
 
     @overload
     @staticmethod
@@ -80,8 +89,7 @@ class injected_self_func(Protocol[T_co, P, R_co]):
 
     @overload
     @staticmethod
-    def __call__(cls: type[T_co], *args: P.args, **kwargs: P.kwargs) -> R_co:
-        ...
+    def __call__(cls: type[T_co], *args: P.args, **kwargs: P.kwargs) -> R_co: ...
 
 
 self_objects_cache = dict[Any, Any]()
@@ -116,7 +124,9 @@ class inject_self_base(Generic[T_co, P, R_co]):
         self.clean_kwargs = False
 
     def __get__(
-        self, class_obj: type[T] | T | None, class_type: type[T] | type[type[T]] | Any  # type: ignore[valid-type]
+        self,
+        class_obj: type[T] | T | None,
+        class_type: type[T | type[T]] | Any,  # type: ignore[valid-type]
     ) -> injected_self_func[T_co, P, R_co]:
         if not self.signature or not self.first_key:
             self.signature = Signature.from_callable(self.function, eval_str=True)
@@ -127,25 +137,19 @@ class inject_self_base(Generic[T_co, P, R_co]):
 
                 if 4 not in {x.kind for x in self.signature.parameters.values()}:
                     raise CustomValueError(
-                        'This function hasn\'t got any kwargs!', 'inject_self.init_kwargs', self.function
+                        "This function hasn't got any kwargs!", "inject_self.init_kwargs", self.function
                     )
 
-                self.init_kwargs = list[str](
-                    k for k, x in self.signature.parameters.items() if x.kind != 4
-                )
+                self.init_kwargs = list[str](k for k, x in self.signature.parameters.items() if x.kind != 4)
 
         @wraps(self.function)
         def _wrapper(*args: Any, **kwargs: Any) -> Any:
-            first_arg = (args[0] if args else None) or (
-                kwargs.get(self.first_key, None) if self.first_key else None
-            )
+            first_arg = (args[0] if args else None) or (kwargs.get(self.first_key, None) if self.first_key else None)
 
-            if (
-                first_arg and (
-                    (is_obj := isinstance(first_arg, class_type))
-                    or isinstance(first_arg, type(class_type))
-                    or first_arg is class_type
-                )
+            if first_arg and (
+                (is_obj := isinstance(first_arg, class_type))
+                or isinstance(first_arg, type(class_type))
+                or first_arg is class_type
             ):
                 obj = first_arg if is_obj else first_arg()
                 if args:
@@ -169,7 +173,7 @@ class inject_self_base(Generic[T_co, P, R_co]):
             else:
                 obj = class_obj
 
-            return self.function(obj, *args, **kwargs)
+            return self.function(obj, *args, **kwargs)  # type: ignore
 
         return _wrapper
 
@@ -191,6 +195,7 @@ class inject_self_base(Generic[T_co, P, R_co]):
             inj.args = args
             inj.kwargs = kwargs
             return inj  # type: ignore
+
         return _wrapper
 
 
@@ -207,9 +212,7 @@ class inject_self(inject_self_base[T_co, P, R_co]):
             def __init__(self, function: Callable[[T1_co], R1_co]) -> None:
                 self.function = inject_self(function)
 
-            def __get__(
-                self, class_obj: type[T1_co] | T1_co | None, class_type: type[T1_co] | T1_co
-            ) -> R1_co:
+            def __get__(self, class_obj: type[T1_co] | T1_co | None, class_type: type[T1_co] | T1_co) -> R1_co:
                 return self.function.__get__(class_obj, class_type)()
 
     class init_kwargs(inject_self_base[T0_co, P0, R0_co]):
@@ -229,9 +232,7 @@ class inject_self(inject_self_base[T_co, P, R_co]):
         def __init__(self, function: Callable[[T0_co], R0_co]) -> None:
             self.function = inject_self(function)
 
-        def __get__(
-            self, class_obj: type[T0_co] | T0_co | None, class_type: type[T0_co] | T0_co
-        ) -> R0_co:
+        def __get__(self, class_obj: type[T0_co] | T0_co | None, class_type: type[T0_co] | T0_co) -> R0_co:
             return self.function.__get__(class_obj, class_type)()
 
 
@@ -243,16 +244,14 @@ class inject_kwargs_params_base_func(Generic[T_co, P, R_co]):
 class inject_kwargs_params_base(Generic[T_co, P, R_co]):
     signature: Signature | None
 
-    _kwargs_name = 'kwargs'
+    _kwargs_name = "kwargs"
 
     def __init__(self, function: Callable[Concatenate[T_co, P], R_co]) -> None:
         self.function = function
 
         self.signature = None
 
-    def __get__(
-        self, class_obj: T, class_type: type[T]
-    ) -> inject_kwargs_params_base_func[T_co, P, R_co]:
+    def __get__(self, class_obj: T, class_type: type[T]) -> inject_kwargs_params_base_func[T_co, P, R_co]:
         if not self.signature:
             self.signature = Signature.from_callable(self.function, eval_str=True)
 
@@ -263,7 +262,7 @@ class inject_kwargs_params_base(Generic[T_co, P, R_co]):
                 from ..exceptions import CustomValueError
 
                 raise CustomValueError(
-                    'This function hasn\'t got any kwargs!', 'inject_kwargs_params.add_to_kwargs', self.function
+                    "This function hasn't got any kwargs!", "inject_kwargs_params.add_to_kwargs", self.function
                 )
 
         this = self
@@ -329,21 +328,20 @@ class inject_kwargs_params_base(Generic[T_co, P, R_co]):
 
 
 if TYPE_CHECKING:  # love you mypy...
+
     class _add_to_kwargs:
-        def __call__(self, func: F1) -> F1:
-            ...
+        def __call__(self, func: F1) -> F1: ...
 
     class _inject_kwargs_params:
-        def __call__(self, func: F0) -> F0:
-            ...
+        def __call__(self, func: F0) -> F0: ...
 
         add_to_kwargs = _add_to_kwargs()
 
     inject_kwargs_params = _inject_kwargs_params()
 else:
+
     class inject_kwargs_params(Generic[T, P, R], inject_kwargs_params_base[T, P, R]):
-        class add_to_kwargs(Generic[T0, P0, R0], inject_kwargs_params_base[T0, P0, R0]):
-            ...
+        class add_to_kwargs(Generic[T0, P0, R0], inject_kwargs_params_base[T0, P0, R0]): ...
 
 
 class complex_hash(Generic[T]):
@@ -356,11 +354,7 @@ class complex_hash(Generic[T]):
     def __new__(cls, class_type: T) -> T:  # type: ignore
         class inner_class_type(class_type):  # type: ignore
             def __hash__(self) -> int:
-                return complex_hash.hash(
-                    self.__class__.__name__, *(
-                        getattr(self, key) for key in self.__annotations__.keys()
-                    )
-                )
+                return complex_hash.hash(self.__class__.__name__, *(getattr(self, key) for key in self.__annotations__))
 
         return inner_class_type  # type: ignore
 
@@ -379,14 +373,11 @@ class complex_hash(Generic[T]):
             try:
                 new_hash = hash(value)
             except TypeError:
-                if isinstance(value, Iterable):
-                    new_hash = complex_hash.hash(*value)
-                else:
-                    new_hash = hash(str(value))
+                new_hash = complex_hash.hash(*value) if isinstance(value, Iterable) else hash(str(value))
 
             values.append(str(new_hash))
 
-        return hash('_'.join(values))
+        return hash("_".join(values))
 
 
 def get_subclasses(family: type[T], exclude: Sequence[type[T]] = []) -> list[type[T]]:
@@ -410,17 +401,15 @@ def get_subclasses(family: type[T], exclude: Sequence[type[T]] = []) -> list[typ
     return list(set(_subclasses(family)))
 
 
-class classproperty(Generic[T, R]):
-    """
-    Make a class property. A combination between classmethod and property.
-    """
-
+class classproperty_base(Generic[T, R_co]):
     __isabstractmethod__: bool = False
 
     def __init__(
         self,
-        fget: Callable[[type[T]], R] | classmethod[T, ..., R],
-        fset: Callable[[type[T], R], None] | classmethod[T, Concatenate[R, ...], None] | None = None,
+        fget: Callable[[type[T]], R_co] | classmethod[T, ..., R_co],
+        fset: Callable[Concatenate[type[T], Any, ...], None]
+        | classmethod[T, Concatenate[Any, ...], None]
+        | None = None,
         fdel: Callable[[type[T]], None] | classmethod[T, ..., None] | None = None,
         doc: str | None = None,
     ) -> None:
@@ -429,6 +418,7 @@ class classproperty(Generic[T, R]):
         self.fdel = self._wrap(fdel) if fdel is not None else fdel
 
         self.__doc__ = doc
+        self.__name__ = self.fget.__name__
 
     def _wrap(self, func: Callable[..., R1] | classmethod[T, P1, R1]) -> classmethod[T, P1, R1]:
         if not isinstance(func, classmethod):
@@ -436,156 +426,216 @@ class classproperty(Generic[T, R]):
 
         return func
 
-    def __get__(self, __obj: T | None, __type: type | None = None) -> R:
-        if __type is None:
-            __type = type(__obj)
+    def __set_name__(self, owner: object, name: str) -> None:
+        self.__name__ = name
 
-        return self.fget.__get__(__obj, __type)()
+    def _get_cache(self, type_: type) -> dict[str, Any]:
+        cache_key = getattr(self, "cache_key")
 
-    def __set__(self, __obj: T, __value: Any) -> None:
+        if not hasattr(type_, cache_key):
+            setattr(type_, cache_key, {})
+
+        return getattr(type_, cache_key)
+
+    def __get__(self, obj: T | None, type_: type | None = None) -> R_co:
+        if type_ is None:
+            type_ = type(obj)
+
+        if not isinstance(self, classproperty.cached):
+            return self.fget.__get__(obj, type_)()
+
+        if self.__name__ in (cache := self._get_cache(type_)):
+            return cache[self.__name__]
+
+        value = self.fget.__get__(obj, type_)()
+        cache[self.__name__] = value
+        return value
+
+    def __set__(self, obj: T, value: Any) -> None:
         if not self.fset:
             raise AttributeError(
-                f'classproperty with getter "{self.__name__}" of "{__obj.__class__.__name__}" object has no setter.'
+                f'classproperty with getter "{self.__name__}" of "{obj.__class__.__name__}" object has no setter.'
             )
 
-        self.fset.__get__(None, type(__obj))(__value)
+        type_ = type(obj)
 
-    def __delete__(self, __obj: T) -> None:
+        if self.__name__ in (cache := self._get_cache(type_)):
+            del cache[self.__name__]
+
+        self.fset.__get__(None, type_)(value)
+
+    def __delete__(self, obj: T) -> None:
         if not self.fdel:
             raise AttributeError(
-                f'classproperty with getter "{self.__name__}" of "{__obj.__class__.__name__}" object has no deleter.'
+                f'classproperty with getter "{self.__name__}" of "{obj.__class__.__name__}" object has no deleter.'
             )
 
-        self.fdel.__get__(None, type(__obj))()
+        type_ = type(obj)
 
-    @property
-    def __name__(self) -> str:
-        return self.fget.__name__
+        if self.__name__ in (cache := self._get_cache(type_)):
+            del cache[self.__name__]
+
+        self.fdel.__get__(None, type_)()
 
 
-class cachedproperty(property, Generic[P, R_co, T, T0, P0]):
+class classproperty(classproperty_base[T, R_co]):
+    """
+    A combination of `classmethod` and `property`.
+    """
+
+    class cached(classproperty_base[T0, R0_co]):
+        """
+        A combination of `classmethod` and `property`.
+
+        The value is computed once and then cached in a dictionary (under `cache_key`)
+        attached to the class type. If a setter or deleter is defined and invoked,
+        the cache is cleared.
+        """
+
+        cache_key = "_jetpt_classproperty_cached"
+
+        @classmethod
+        def clear_cache(cls, type_: type, names: str | Iterable[str] | None = None) -> None:
+            """
+            Clear cached properties of an type instance.
+
+            :param type_:   The type whose cache should be cleared.
+            :param names:   Specific property names to clear. If None, all cached properties are cleared.
+            """
+            if names is None:
+                with suppress(AttributeError):
+                    getattr(type_, cls.cache_key).clear()
+                return None
+
+            from ..functions import to_arr
+
+            cache = getattr(type_, cls.cache_key, {})
+
+            for name in to_arr(names):
+                with suppress(KeyError):
+                    del cache[name]
+
+
+class cachedproperty(property, Generic[R_co]):
     """
     Wrapper for a one-time get property, that will be cached.
 
-    Keep in mind two things:
-
-     * The cache is per-object. Don't hold a reference to itself or it will never get garbage collected.
-     * Your class has to either manually set __dict__[cachedproperty.cache_key]
-       or inherit from cachedproperty.baseclass.
+    You shouldn't hold a reference to itself or it will never get garbage collected.
     """
 
     __isabstractmethod__: bool = False
 
-    cache_key = '_jetpt_cachedproperty_cache'
+    cache_key = "_jetpt_cachedproperty_cache"
 
+    @deprecated(
+        "The cache dict is now set automatically. You no longer need to inherit from it", category=DeprecationWarning
+    )
     class baseclass:
         """Inherit from this class to automatically set the cache dict."""
 
-        if not TYPE_CHECKING:
-            def __new__(cls, *args: Any, **kwargs: Any) -> None:
-                try:
-                    self = super().__new__(cls, *args, **kwargs)
-                except TypeError:
-                    self = super().__new__(cls)
-                self.__dict__.__setitem__(cachedproperty.cache_key, dict[str, Any]())
-                return self
-
     if TYPE_CHECKING:
+
         def __init__(
-            self, fget: Callable[P, R_co], fset: Callable[[T, T0], None] | None = None,
-            fdel: Callable[P0, None] | None = None, doc: str | None = None,
-        ) -> None:
-            ...
+            self,
+            fget: Callable[[Any], R_co],
+            fset: Callable[[Any, Any], None] | None = None,
+            fdel: Callable[[Any], None] | None = None,
+            doc: str | None = None,
+        ) -> None: ...
 
-        def getter(self, __fget: Callable[P1, R0_co]) -> cachedproperty[P1, R0_co, T, T0, P0]:
-            ...
+        def getter(self, fget: Callable[..., R_co]) -> cachedproperty[R_co]: ...
 
-        def setter(self, __fset: Callable[[T1, T2], None]) -> cachedproperty[P, R_co, T1, T2, P0]:
-            ...
+        def setter(self, fset: Callable[[Any, Any], None]) -> cachedproperty[R_co]: ...
 
-        def deleter(self, __fdel: Callable[P1, None]) -> cachedproperty[P, R_co, T, T0, P1]:
-            ...
+        def deleter(self, fdel: Callable[..., None]) -> cachedproperty[R_co]: ...
 
     @overload
-    def __get__(self, __obj: None, __type: type | None = None) -> Self:
-        ...
+    def __get__(self, instance: None, owner: type | None = None) -> Self: ...
 
     @overload
-    def __get__(self, __obj: object, __type: type | None = None) -> R_co:
-        ...
+    def __get__(self, instance: Any, owner: type | None = None) -> R_co: ...
 
-    def __get__(self, __obj: Any, __type: type | None = None) -> Any:
-        if isinstance(self.fget, classproperty):
-            function = partial(self.fget.__get__, __obj, __type)  # type: ignore
-            __obj = __type
+    def __get__(self, instance: Any, owner: type | None = None) -> Any:
+        if instance is None:
+            return self
 
-            if not hasattr(__obj, cachedproperty.cache_key):
-                setattr(__obj, cachedproperty.cache_key, dict[str, Any]())
+        if self.__name__ in (cache := instance.__dict__.setdefault(self.cache_key, {})):
+            return cache[self.__name__]
 
-            cache = getattr(__obj, cachedproperty.cache_key)
-            name = self.fget.__name__
-        else:
-            assert self.fget
-            function = self.fget.__get__(__obj, __type)
-            cache = __obj.__dict__.get(cachedproperty.cache_key)
-            name = function.__name__
+        value = super().__get__(instance, owner)
+        cache[self.__name__] = value
+        return value
 
-        if name not in cache:
-            cache[name] = function()
+    def __set__(self, instance: Any, value: Any) -> None:
+        if self.__name__ in (cache := instance.__dict__.setdefault(self.cache_key, {})):
+            del cache[self.__name__]
 
-        return cache[name]
+        return super().__set__(instance, value)
 
-    if TYPE_CHECKING:
-        def __set__(self, obj: Any, value: R_co, /) -> None:  # type: ignore[misc]
-            ...
+    def __delete__(self, instance: Any) -> None:
+        if self.__name__ in (cache := instance.__dict__.setdefault(self.cache_key, {})):
+            del cache[self.__name__]
+
+        return super().__delete__(instance)
+
+    @classmethod
+    def clear_cache(cls, obj: object, names: str | Iterable[str] | None = None) -> None:
+        """
+        Clear cached properties of an object instance.
+
+        :param obj:   The object whose cache should be cleared.
+        :param names: Specific property names to clear. If None, all cached properties are cleared.
+        """
+        if names is None:
+            obj.__dict__.get(cls.cache_key, {}).clear()
+            return None
+
+        from ..functions import to_arr
+
+        cache = obj.__dict__.get(cls.cache_key, {})
+
+        for name in to_arr(names):
+            with suppress(KeyError):
+                del cache[name]
 
 
 class KwargsNotNone(KwargsT):
     """Remove all None objects from this kwargs dict."""
 
     if not TYPE_CHECKING:
-        def __new__(cls, *args: Any, **kwargs: Any) -> KwargsNotNone:
-            return KwargsT(**{
-                key: value for key, value in KwargsT(*args, **kwargs).items()
-                if value is not None
-            })
 
-
-SingleMeta = TypeVar('SingleMeta', bound=type)
+        def __new__(cls, *args: Any, **kwargs: Any) -> Self:
+            return KwargsT(**{key: value for key, value in KwargsT(*args, **kwargs).items() if value is not None})
 
 
 class SingletonMeta(type):
-    _instances = dict[type[SingleMeta], SingleMeta]()  # type: ignore
+    _instances: ClassVar[dict[type[Any], Any]] = {}
     _singleton_init: bool
 
-    def __new__(
-        cls: type[SingletonSelf], name: str, bases: tuple[type, ...], namespace: dict[str, Any], **kwargs: Any
-    ) -> SingletonSelf:
-        return type.__new__(cls, name, bases, namespace | {'_singleton_init': kwargs.pop('init', False)})
+    def __new__(cls, name: str, bases: tuple[type, ...], namespace: dict[str, Any], **kwargs: Any) -> SingletonMeta:
+        return type.__new__(cls, name, bases, namespace | {"_singleton_init": kwargs.pop("init", False)})
 
-    def __call__(cls: type[SingletonSelf], *args: Any, **kwargs: Any) -> SingletonSelf:  # type: ignore
+    def __call__(cls, *args: Any, **kwargs: Any) -> SingletonMeta:
         if cls not in cls._instances:
             cls._instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
         elif cls._singleton_init:
-            cls._instances[cls].__init__(*args, **kwargs)  # type: ignore
+            cls._instances[cls].__init__(*args, **kwargs)
 
         return cls._instances[cls]
 
 
-SingletonSelf = TypeVar('SingletonSelf', bound=SingletonMeta)
-
-
 class Singleton(metaclass=SingletonMeta):
     """Handy class to inherit to have the SingletonMeta metaclass."""
 
 
 class to_singleton_impl:
     _ts_args = tuple[str, ...]()
-    _ts_kwargs = dict[str, Any]()
+    _ts_kwargs: Mapping[str, Any] = {}
     _add_classes = tuple[type, ...]()
 
     def __new__(_cls, cls: type[T]) -> T:  # type: ignore
         if _cls._add_classes:
+
             class rcls(cls, *_cls._add_classes):  # type: ignore
                 ...
         else:
@@ -604,11 +654,11 @@ class to_singleton_impl:
 
 class to_singleton(to_singleton_impl):
     class as_property(to_singleton_impl):
-        _add_classes = (property, )
+        _add_classes = (property,)
 
 
 class LinearRangeLut(Mapping[int, int]):
-    __slots__ = ('ranges', '_ranges_idx_lut', '_misses_n')
+    __slots__ = ("_misses_n", "_ranges_idx_lut", "ranges")
 
     def __init__(self, ranges: Mapping[int, range]) -> None:
         self.ranges = ranges
@@ -617,6 +667,8 @@ class LinearRangeLut(Mapping[int, int]):
         self._misses_n = 0
 
     def __getitem__(self, n: int) -> int:
+        missed_hit = 0
+
         for missed_hit, (idx, k) in enumerate(self._ranges_idx_lut):
             if n in k:
                 break
